/* -*- Mode: Javascript; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Mozilla Penelope project.
 *
 * The Initial Developer of the Original Code is
 * QUALCOMM Incorporated.
 * Portions created by QUALCOMM Incorporated are
 * Copyright (C) 2007-2010 QUALCOMM Incorporated. All Rights Reserved.
 *
 * Contributor(s):
 *     Mark Charlebois <mcharleb@qualcomm.com> original author
 *     Jeff Beckley <beckley@qualcomm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const kRowMax = 3;
const kVSizeSlop = 5;

var gToolboxDocument = null;
var gToolbox = null;
var gCurrentDragOverItem = null;
var gToolboxChanged = false;
var gToolboxIconSize = false;
var gToolboxSheet = false;
var gPreviousMode = null;
var gPreviousIconSize = null;

var gMboxTree = null;
var gMenuItemTree = null;
var gRecipientTree = null;

function onLoad()
{
  if ("arguments" in window && window.arguments[0]) {
    InitWithToolbox(window.arguments[0]);
    repositionDialog();
  }
  else if (window.frameElement &&
           "toolbox" in window.frameElement) {
    gToolboxSheet = true;
    InitWithToolbox(window.frameElement.toolbox);
  }
}

function InitWithToolbox(aToolbox)
{
  gToolbox = aToolbox;
  gToolboxDocument = gToolbox.ownerDocument;
  gToolbox.customizing = true;

  gToolbox.addEventListener("dragstart", onToolbarDragStart, false);
  gToolbox.addEventListener("dragover", onToolbarDragOver, false);
  gToolbox.addEventListener("dragleave", onToolbarDragLeave, false);
  gToolbox.addEventListener("drop", onToolbarDrop, false);

  initDialog();
}

function onClose()
{
  if (!gToolboxSheet)
    window.close();
  else
    finishToolbarCustomization();
}

function onUnload()
{
  if (!gToolboxSheet)
    finishToolbarCustomization();
}

function finishToolbarCustomization()
{
  gToolbox.customizing = false;
  removeToolboxListeners();
  unwrapToolbarItems(true);
  persistCurrentSets();

  notifyParentComplete();
}

function onCancel()
{
  // restore the saved toolbarset for each customizeable toolbar
  unwrapToolbarItems(false);
  var toolbar = gToolbox.firstChild;
  while (toolbar) {
    if (isCustomizableToolbar(toolbar)) {
      if (!toolbar.hasAttribute("customindex")) {
        var previousset = toolbar.getAttribute("previousset");
        if (previousset)
        {
          toolbar.currentSet = previousset;
        }
      }
    }
    toolbar = toolbar.nextSibling;
  }

  // restore the previous mode and iconsize
  updateIconSize(gPreviousIconSize == "small", false);
  updateToolbarMode(gPreviousMode, false);

  repositionDialog();
  gToolboxChanged = true;

  return onAccept(); // we restored the default toolbar, act like a normal accept event now
}

function onAccept(aEvent)
{
  if (navigator.platform.match("^Mac"))
    finishToolbarCustomization();
  else
    window.close();
}

function initDialog()
{
  // Get mode from toolbar first, then toolbox default value if not found
  try {
    gPreviousMode = gToolbox.firstChild.getAttribute("mode");
  }
  catch (e) {
    gPreviousMode = gToolbox.getAttribute("mode");
  }
  document.getElementById("modelist").value = gPreviousMode;
  
  // Get iconsize from toolbar first, then toolbox default value if not found
  try {
    gToolboxIconSize = gPreviousIconSize = gToolbox.firstChild.getAttribute("iconsize");
  }
  catch (e) {
    gToolboxIconSize = gPreviousIconSize = gToolbox.getAttribute("iconsize");
  }
  gToolboxIconSize = gPreviousIconSize = gToolbox.getAttribute("iconsize");
  var smallIconsCheckbox = document.getElementById("smallicons");
  if (gPreviousMode == "text")
    smallIconsCheckbox.disabled = true;
  else
    smallIconsCheckbox.checked = gPreviousIconSize == "small";

  // Build up the palette of other items.
  buildFormatItemPalette();
  buildExtrasPalette();

  // Wrap all the items on the toolbar in toolbarpaletteitems.
  wrapToolbarItems(true);

  gMenuItemTree = document.getElementById("menubartree");
  gMboxTree = document.getElementById("mailboxtree");
  gRecipientTree = document.getElementById("abResultsTree");
  var menu = gToolboxDocument.getElementsByTagName("menubar").item(0);

  createMenubarTree(menu, gMenuItemTree.getElementsByTagName("treechildren").item(0));

  try {
    // Select the first item of the various trees
    gMenuItemTree.view.selection.select(0);
    gMboxTree.view.selection.select(0);
  } catch (e) {
    dump("Couldn't select first tree items: "+e+"\n");
  }

  // from abcommon.js
  InitCommonJS();
  gSearchInput = document.getElementById("peopleSearchInput");

  // Select personal addresses directory
  var treeBuilder = dirTree.builder.QueryInterface(Components.interfaces.nsIXULTreeBuilder);
  SelectFirstAddressBook();

  //SetupAbCommandUpdateHandlers();
}

function repositionDialog()
{
  var win = document.lastChild;

  // Position the dialog touching the bottom of the toolbox and centered with 
  // it. We must resize the window smaller first so that it is positioned 
  // properly. 
  var screenX = gToolbox.boxObject.screenX + ((gToolbox.boxObject.width - win.width) / 2);
  var screenY = gToolbox.boxObject.screenY + gToolbox.boxObject.height;

  var newHeight = win.height;
  if (newHeight >= screen.availHeight - screenY - kVSizeSlop) {
    newHeight = screen.availHeight - screenY - kVSizeSlop;
  }

  window.resizeTo(win.width, newHeight);
  window.moveTo(screenX, screenY);
}

function removeToolboxListeners()
{
  gToolbox.removeEventListener("dragstart", onToolbarDragStart, false);
  gToolbox.removeEventListener("dragover", onToolbarDragOver, false);
  gToolbox.removeEventListener("dragleave", onToolbarDragLeave, false);
  gToolbox.removeEventListener("drop", onToolbarDrop, false);
}

/**
 * Invoke a callback on the toolbox to notify it that the dialog is done
 * and going away.
 */
function notifyParentComplete()
{
  if ("customizeDone" in gToolbox)
    gToolbox.customizeDone(gToolboxChanged);
}

function toolboxChanged(refreshPalettes)
{
  gToolboxChanged = true;
  if ("customizeChange" in gToolbox)
    gToolbox.customizeChange();

  if (refreshPalettes)
  {
    // An item dragged on/off the toolbar be an may item that is currently
    // being displayed, so refresh the various palettes that need it
    onMenuItemSelect();
    buildExtrasPalette();
  }

  repositionDialog();
}

function getToolbarAt(i)
{
  return gToolbox.childNodes[i];
}

/**
 * Persist the current set of buttons in all customizable toolbars to
 * localstore.
 */
function persistCurrentSets()
{
  if (!gToolboxChanged || gToolboxDocument.defaultView.closed)
   return;

  var customCount = 0;
  for (var i = 0; i < gToolbox.childNodes.length; ++i) {
    // Look for customizable toolbars that need to be persisted.
    var toolbar = getToolbarAt(i);
    if (isCustomizableToolbar(toolbar)) {
      // Calculate currentset and store it in the attribute.
      var currentSet = toolbar.currentSet;
      toolbar.setAttribute("currentset", currentSet);

      var customIndex = toolbar.hasAttribute("customindex");
      if (customIndex) {
        if (!toolbar.firstChild) {
          // Remove custom toolbars whose contents have been removed.
          gToolbox.removeChild(toolbar);
          --i;
        } else {
          // Persist custom toolbar info on the <toolbarset/>
          gToolbox.toolbarset.setAttribute("toolbar"+(++customCount),
                                           toolbar.toolbarName + ":" + currentSet);
          gToolboxDocument.persist(gToolbox.toolbarset.id, "toolbar"+customCount);
        }
      }

      if (!customIndex) {
        // Persist the currentset attribute directly on hardcoded toolbars.
        gToolboxDocument.persist(toolbar.id, "currentset");
      }
    }
  }

  // Remove toolbarX attributes for removed toolbars.
  while (gToolbox.toolbarset.hasAttribute("toolbar"+(++customCount))) {
    gToolbox.toolbarset.removeAttribute("toolbar"+customCount);
    gToolboxDocument.persist(gToolbox.toolbarset.id, "toolbar"+customCount);
  }
}

/**
 * Wraps all items in all customizable toolbars in a toolbox.
 */
function wrapToolbarItems(aStorePreviousSet)
{
  for (var i = 0; i < gToolbox.childNodes.length; ++i) {
    var toolbar = getToolbarAt(i);
    if (isCustomizableToolbar(toolbar)) {
      if (aStorePreviousSet)
        toolbar.setAttribute('previousset', toolbar.currentSet);

      for (var k = 0; k < toolbar.childNodes.length; ++k) {
        var item = toolbar.childNodes[k];

        if (navigator.platform.match("^Mac")) {
          if (item.firstChild && item.firstChild.localName == "menubar")
            continue;
        }

        if (isToolbarItem(item)) {
          var wrapper = wrapToolbarItem(item);
          cleanupItemForToolbar(item, wrapper);
        }
      }
    }
  }
}

/**
 * Unwraps all items in all customizable toolbars in a toolbox.
 */
 function unwrapToolbarItems(aRemovePreviousSet)
{
  for (var i = 0; i < gToolbox.childNodes.length; ++i) {
    var toolbar = getToolbarAt(i);
    if (isCustomizableToolbar(toolbar)) {
      if (aRemovePreviousSet)
        toolbar.removeAttribute('previousset');

      for (var k = 0; k < toolbar.childNodes.length; ++k) {
        var paletteItem = toolbar.childNodes[k];
        var toolbarItem = paletteItem.firstChild;

        if (toolbarItem && isToolbarItem(toolbarItem)) {
          var nextSibling = paletteItem.nextSibling;

          if (paletteItem.hasAttribute("itemcommand"))
            toolbarItem.setAttribute("command", paletteItem.getAttribute("itemcommand"));
          if (paletteItem.hasAttribute("itemobserves"))
            toolbarItem.setAttribute("observes", paletteItem.getAttribute("itemobserves"));

          paletteItem.removeChild(toolbarItem); 
          paletteItem.parentNode.removeChild(paletteItem);

          if (nextSibling)
            toolbar.insertBefore(toolbarItem, nextSibling);
          else
            toolbar.appendChild(toolbarItem);
        }
      }
    }
  }
}


/**
 * Creates a wrapper that can be used to contain a toolbaritem and prevent
 * it from receiving UI events.
 */
function createWrapper(aId, aDocument)
{
  var wrapper = aDocument.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                         "toolbarpaletteitem");

  wrapper.id = "wrapper-"+aId;  
  return wrapper;
}

/**
 * Wraps an item that has been cloned from a template and adds
 * it to the end of a row in the palette.
 */
function wrapPaletteItem(aPaletteItem, aCurrentRow, aSpacer)
{
  var wrapper = createWrapper(aPaletteItem.id, document);

  wrapper.setAttribute("flex", 1);
  wrapper.setAttribute("align", "center");
  wrapper.setAttribute("pack", "center");
  wrapper.setAttribute("minheight", "0");
  wrapper.setAttribute("minwidth", "0");

  wrapper.appendChild(aPaletteItem);

  // XXX We need to call this AFTER the palette item has been appended
  // to the wrapper or else we crash dropping certain buttons on the 
  // palette due to removal of the command and disabled attributes - JRH
  cleanUpItemForPalette(aPaletteItem, wrapper);

  if (aSpacer)
    aCurrentRow.insertBefore(wrapper, aSpacer);
  else
    aCurrentRow.appendChild(wrapper);

}

/**
 * Wraps an item that is currently on a toolbar and replaces the item
 * with the wrapper. This is not used when dropping items from the palette,
 * only when first starting the dialog and wrapping everything on the toolbars.
 */
function wrapToolbarItem(aToolbarItem)
{
  var wrapper = createWrapper(aToolbarItem.id, gToolboxDocument);

  cleanupItemForToolbar(aToolbarItem, wrapper);
  wrapper.flex = aToolbarItem.flex;

  aToolbarItem.parentNode.replaceChild(wrapper, aToolbarItem);
  
  wrapper.appendChild(aToolbarItem);
  
  return wrapper;
}

/**
 * Get the list of ids for the current set of items on each toolbar.
 */
function getCurrentItemIds()
{
  var currentItems = {};
  for (var i = 0; i < gToolbox.childNodes.length; ++i) {
    var toolbar = getToolbarAt(i);
    if (isCustomizableToolbar(toolbar)) {
      var child = toolbar.firstChild;
      while (child) {
        if (isToolbarItem(child))
          currentItems[child.id] = 1;
        child = child.nextSibling;
      }
    }
  }
  return currentItems;
}

/**
 * Builds the palette of draggable items that are not yet in a toolbar.
 */
function buildPalette(mypalette, toolbaritem, descId)
{
  // Empty the palette first.
  var paletteBox = document.getElementById(mypalette);
  while (paletteBox.lastChild)
    paletteBox.removeChild(paletteBox.lastChild);

  var currentRow = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                            "hbox");
  currentRow.setAttribute("class", "paletteRow");

  var rowSlot = 0;

  var items = gToolbox.palette.getElementsByAttribute('id', toolbaritem);
  if (items.length) {
    var paletteItem = document.importNode(items.item(0), true);

    wrapPaletteItem(paletteItem, currentRow, null);
    setDescText(descId, paletteItem.getAttribute("tooltiptext"));
  } else {
    notSupported(mypalette, descId);
  }

  if (currentRow) { 
    fillRowWithFlex(currentRow);
    paletteBox.appendChild(currentRow);
  }
}

/**
 * Builds the palette of draggable items that are not yet in a toolbar.
 */
function buildFormatItemPalette()
{
  var descId = "formatItemDescription";

  // Empty the palette first.
  var paletteBox = document.getElementById("format-palette");
  while (paletteBox.lastChild)
    paletteBox.removeChild(paletteBox.lastChild);

  var currentRow = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                            "hbox");
  currentRow.setAttribute("class", "paletteRow");

  // Add the toolbar separator item.
  var templateNode = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                              "toolbarseparator");
  templateNode.id = "separator";
  wrapPaletteItem(templateNode, currentRow, null);

  // Add the toolbar spring item.
  templateNode = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                          "toolbarspring");
  templateNode.id = "spring";
  templateNode.flex = 1;
  wrapPaletteItem(templateNode, currentRow, null);

  // Add the toolbar spacer item.
  templateNode = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                          "toolbarspacer");
  templateNode.id = "spacer";
  templateNode.flex = 1;
  wrapPaletteItem(templateNode, currentRow, null);

  var stringBundlePenelope = document.getElementById("stringBundlePenelope");
  var desc = stringBundlePenelope.getString("formattingDescription");
  setDescText(descId, desc);

  if (currentRow) { 
    fillRowWithFlex(currentRow);
    paletteBox.appendChild(currentRow);
  }
}

/**
 * Creates a new palette item for a cloned template node and
 * adds it to the last slot in the palette.
 */
function appendPaletteItem(aItem, aPalette)
{
  var lastRow = aPalette.lastChild;
  var lastSpacer = lastRow.lastChild;
 
  if (lastSpacer.localName != "spacer") {
    // The current row is full, so we have to create a new row.
    lastRow = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                       "hbox");
    lastRow.setAttribute("class", "paletteRow");
    aPalette.appendChild(lastRow);
  
    wrapPaletteItem(aItem, lastRow, null);

    fillRowWithFlex(lastRow);
  } else {
    // Decrement the flex of the last spacer or remove it entirely.
    var flex = lastSpacer.getAttribute("flex");
    if (flex == 1) {
      lastRow.removeChild(lastSpacer);
      lastSpacer = null;
    } else
      lastSpacer.setAttribute("flex", --flex);

    // Insert the wrapper where the last spacer was.
    wrapPaletteItem(aItem, lastRow, lastSpacer);
  }
}

/**
 * Builds the palette of draggable items that are not yet in a toolbar.
 */
function buildMailboxPalette(uri, name, disableMoveto)
{
  var descId = "mailboxDescription";

  // Empty the palette first.
  var paletteBox = document.getElementById("mailbox-palette");
  while (paletteBox.lastChild)
    paletteBox.removeChild(paletteBox.lastChild);

  var currentRow = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                            "hbox");
  currentRow.setAttribute("class", "paletteRow");

  var paletteItem = createFolderOpenButton(uri, name);
  wrapPaletteItem(paletteItem, currentRow, null);

  var stringBundlePenelope = document.getElementById("stringBundlePenelope");
  if (!disableMoveto) {
    var paletteItem = createFolderMovetoButton(uri, name);
    wrapPaletteItem(paletteItem, currentRow, null);

    var desc = stringBundlePenelope.getFormattedString("openTransferMailboxDescription", [name]);
    setDescText(descId, desc);
  } else {
    var desc = stringBundlePenelope.getFormattedString("openMailboxDescription", [name]);
    setDescText(descId, desc);
  }

  if (currentRow) { 
    fillRowWithFlex(currentRow);
    paletteBox.appendChild(currentRow);
  }
}

/**
 * Builds the palette of draggable items for recipients.
 */
function buildRecipientsPalette(email, name)
{
  var descId = "recipientDescription";

  // Empty the palette first.
  var paletteBox = document.getElementById("recipient-palette");
  while (paletteBox.lastChild)
    paletteBox.removeChild(paletteBox.lastChild);

  var currentRow = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                            "hbox");
  currentRow.setAttribute("class", "paletteRow");

  var node = gToolbox.palette.firstChild;
  var rowSlot = 1;

  // Add the create new message item.
  var newMailNode = createRecipientButton(email, name, "p-recipient-");
  newMailNode.flex = 1;
  wrapPaletteItem(newMailNode, currentRow, null);

  var forwardNode = createRecipientButton(email, name, "p-forward-");
  forwardNode.flex = 1;
  wrapPaletteItem(forwardNode, currentRow, null);

  var redirectNode = createRecipientButton(email, name, "p-redirect-");
  redirectNode.flex = 1;
  wrapPaletteItem(redirectNode, currentRow, null);

  // Not adding Insert recipient text as it's not implemented
  //var insertNode = createRecipientButton(email, name, "p-insert-");
  //insertNode.flex = 1;
  //wrapPaletteItem(insertNode, currentRow, null);

  var stringBundlePenelope = document.getElementById("stringBundlePenelope");
  var desc = stringBundlePenelope.getString("recipientDescription");
  setDescText(descId, desc);

  if (currentRow) { 
    fillRowWithFlex(currentRow);
    paletteBox.appendChild(currentRow);
  }
}

/**
 * Builds the palette of draggable items that are not yet in a toolbar.
 */
function buildExtrasPalette()
{
  // Empty the palette first.
  var paletteBox = document.getElementById("extras-palette");
  while (paletteBox.lastChild)
    paletteBox.removeChild(paletteBox.lastChild);

  var currentRow = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                            "hbox");
  currentRow.setAttribute("class", "paletteRow");

  var node = gToolbox.palette.firstChild;
  var rowSlot = 1;
  var ignore;
  for (; node; node = node.nextSibling) {
    if (isSpecialItem(node))
      continue;

    ignore = true;
    switch(node.id) {
    case "button-getmsg":
    case "button-newmsg":
    case "button-reply":
    case "button-replyall":
    case "button-forward":
    case "button-file":
    case "button-goback":
    case "button-goforward":
    case "button-previous":
    case "button-next":
    case "button-junk":
    case "button-delete":
    case "button-mark":
    case "button-tag":
    case "button-address":
    case "throbber-box":
    case "button-stop":
    case "button-in":
    case "button-out":
    case "button-help":
    case "button-newcard":
    case "button-opensavedmessage":
    case "button-close":
    case "button-saveas":
    case "button-subscribe":
    case "button-compact":
    case "button-emptytrash":
    case "button-offline":
    case "button-printsetup":
    case "button-printpreview":
    case "button-print":
    case "button-quit":
    case "cut-button":
    case "copy-button":
    case "paste-button":
    case "undo-button":
    case "redo-button":
    case "button-selectall":
    case "button-find":
    case "button-findagain":
    case "button-search":
    case "button-properties":
    case "mailviews-container":
    case "button-viewpagesource":
    case "button-viewsecuritystatus":
    case "button-archive":
    case "button-redirect":
    case "button-resend":
    case "button-sendunsent":
    case "button-nextunread":
    case "button-nextstarred":
    case "button-nextunreadthread":
    case "button-prevunread":
    case "button-prevstarred":
    case "button-prefs":
    case "button-addtoaddrbook":
    case "button-addons":
    case "button-filters":
    case "button-applyfilters":
    case "button-applyfilterstoselection":
    case "button-createfilter":
    case "button-import":
    case "priority-button":
    case "button-status":
      break;
    default:
      if ((node.id.substring(0,7) == "p-open-") ||
          (node.id.substring(0,9) == "p-moveto-") ||
          (node.id.substring(0,12) == "p-recipient-") ||
          (node.id.substring(0,10) == "p-forward-") ||
          (node.id.substring(0,11) == "p-redirect-") ||
          (node.id.substring(0,9)  == "p-insert-")) {
      } else {
        ignore = false;
      }
    }
    if (!ignore) {
      dump("Extra toolbar item: "+node.id+"\n");
      var paletteItem = document.importNode(node, true);
      if (rowSlot == kRowMax) {
        // Append the old row.
        paletteBox.appendChild(currentRow);

        // Make a new row.
        currentRow = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                              "hbox");
        currentRow.setAttribute("class", "paletteRow");
        rowSlot = 0;
      }
      ++rowSlot;
      wrapPaletteItem(paletteItem, currentRow, null);
    }
  }

  if (currentRow) { 
    fillRowWithFlex(currentRow);
    paletteBox.appendChild(currentRow);
  }
}

function fillRowWithFlex(aRow)
{
  var remainingFlex = kRowMax - aRow.childNodes.length;
  if (remainingFlex > 0) {
    var spacer = document.createElementNS("http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul",
                                          "spacer");
    spacer.setAttribute("flex", remainingFlex);
    aRow.appendChild(spacer);
  }
}

/**
 * Makes sure that an item that has been cloned from a template
 * is stripped of all properties that may adversely affect its
 * appearance in the palette.
 */
function cleanUpItemForPalette(aItem, aWrapper)
{
  aWrapper.setAttribute("place", "palette");
  setWrapperType(aItem, aWrapper);

  if (aItem.hasAttribute("title"))
    aWrapper.setAttribute("title", aItem.getAttribute("title"));
  else if (isSpecialItem(aItem)) {
    var stringBundle = document.getElementById("stringBundle");
    // Remove the common "toolbar" prefix to generate the string name.
    var title = stringBundle.getString(aItem.localName.slice(7) + "Title");
    aWrapper.setAttribute("title", title);
  }

  // Remove attributes that screw up our appearance.
  aItem.removeAttribute("command");
  aItem.removeAttribute("observes");
  aItem.removeAttribute("disabled");
  aItem.removeAttribute("type");
  aItem.removeAttribute("width");

  if (aItem.localName == "toolbaritem" && aItem.firstChild) {
    aItem.firstChild.removeAttribute("observes");

    // So the throbber doesn't throb in the dialog,
    // cute as that may be...
    aItem.firstChild.removeAttribute("busy");
  }
}

/**
 * Makes sure that an item that has been cloned from a template
 * is stripped of all properties that may adversely affect its
 * appearance in the toolbar.  Store critical properties on the 
 * wrapper so they can be put back on the item when we're done.
 */
function cleanupItemForToolbar(aItem, aWrapper)
{
  setWrapperType(aItem, aWrapper);
  aWrapper.setAttribute("place", "toolbar");

  if (aItem.hasAttribute("command")) {
    aWrapper.setAttribute("itemcommand", aItem.getAttribute("command"));
    aItem.removeAttribute("command");
  }

  if (aItem.hasAttribute("observes"))
  {
    aWrapper.setAttribute("itemobserves", aItem.getAttribute("observes"));
    aItem.removeAttribute("observes");
  }

  if (aItem.disabled) {
    aWrapper.setAttribute("itemdisabled", "true");
    aItem.disabled = false;
  }
}

function setWrapperType(aItem, aWrapper)
{
  try {
    if (aItem.localName == "toolbarseparator") {
      aWrapper.setAttribute("type", "separator");
    } else if (aItem.localName == "toolbarspring") {
      aWrapper.setAttribute("type", "spring");
    } else if (aItem.localName == "toolbarspacer") {
      aWrapper.setAttribute("type", "spacer");
    } else if (aItem.localName == "toolbaritem" && aItem.firstChild) {
      aWrapper.setAttribute("type", aItem.firstChild.localName);
    }
  }
  catch(e) {
    dump("Exception: aItem = "+aItem+"\n");
  }
}

function setDragActive(aItem, aValue)
{
  var node = aItem;
  var direction = window.getComputedStyle(aItem, null).direction;
  var value = direction == "ltr"? "left" : "right";
  if (aItem.localName == "toolbar") {
    node = aItem.lastChild;
    value = direction == "ltr"? "right" : "left";
  }

  if (!node)
    return;

  if (aValue) {
    if (!node.hasAttribute("dragover"))
      node.setAttribute("dragover", value);
  } else {
    node.removeAttribute("dragover");
  }
}

function addNewToolbar()
{
  var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                                .getService(Components.interfaces.nsIPromptService);

  var stringBundle = document.getElementById("stringBundle");
  var message = stringBundle.getString("enterToolbarName");
  var title = stringBundle.getString("enterToolbarTitle");

  var name = {};

  while (true) {

    if (!promptService.prompt(window, title, message, name, null, {})) {
      return;
    }
    
    if (!name.value) {
      message = stringBundle.getFormattedString("enterToolbarBlank", [name.value]);
      continue;
    }

    var dupeFound = false;

     // Check for an existing toolbar with the same display name
    for (i = 0; i < gToolbox.childNodes.length; ++i) {
      var toolbar = gToolbox.childNodes[i];
      var toolbarName = toolbar.getAttribute("toolbarname");

      if (toolbarName == name.value &&
          toolbar.getAttribute("type") != "menubar" &&
          toolbar.nodeName == 'toolbar') {
        dupeFound = true;
        break;
      }
    }

    if (!dupeFound)
      break;

    message = stringBundle.getFormattedString("enterToolbarDup", [name.value]);
  }
    
  gToolbox.appendCustomToolbar(name.value, "");
  
  toolboxChanged();
}

/**
 * Restore the default set of buttons to fixed toolbars,
 * remove all custom toolbars, and rebuild the palette.
 */
function restoreDefaultSet()
{
  // Restore the defaultset for fixed toolbars.
  unwrapToolbarItems(false);
  var toolbar = gToolbox.firstChild;
  while (toolbar) {
    if (isCustomizableToolbar(toolbar)) {
      if (!toolbar.hasAttribute("customindex")) {
        var defaultSet = toolbar.getAttribute("defaultset");
        if (defaultSet)
        {
          toolbar.currentSet = defaultSet;
        }
      }
    }
    toolbar = toolbar.nextSibling;
  }

  // Restore the default icon size (large) and mode (just icons).
  updateIconSize(false, false);
  document.getElementById("smallicons").checked = false;
  updateToolbarMode("icons", false);
  document.getElementById("modelist").value = "icons";

  // Remove all of the customized toolbars.
  var child = gToolbox.lastChild;
  while (child) {
    if (child.hasAttribute("customindex")) {
      var thisChild = child;
      child = child.previousSibling;
      thisChild.currentSet = "__empty";
      gToolbox.removeChild(thisChild);
    } else {
      child = child.previousSibling;
    }
  }

  // rebuild up the palette of other items.
  buildFormatItemPalette();
  buildExtrasPalette();

  // Now re-wrap the items on the toolbar, but don't clobber previousset.
  wrapToolbarItems(false);

  toolboxChanged();
}

function updateIconSize(aUseSmallIcons, localDefault)
{
  gToolboxIconSize = aUseSmallIcons ? "small" : "large";
  
  setAttribute(gToolbox, "iconsize", gToolboxIconSize);
  gToolboxDocument.persist(gToolbox.id, "iconsize");
  
  for (var i = 0; i < gToolbox.childNodes.length; ++i) {
    var toolbar = getToolbarAt(i);
    if (isCustomizableToolbar(toolbar)) {
      var toolbarIconSize = (localDefault && toolbar.hasAttribute("defaulticonsize")) ?
                            toolbar.getAttribute("defaulticonsize") :
                            gToolboxIconSize;
      setAttribute(toolbar, "iconsize", toolbarIconSize);
      gToolboxDocument.persist(toolbar.id, "iconsize");
    }
  }

  toolboxChanged();
}

function updateToolbarMode(aModeValue, localDefault)
{
  setAttribute(gToolbox, "mode", aModeValue);
  gToolboxDocument.persist(gToolbox.id, "mode");

  for (var i = 0; i < gToolbox.childNodes.length; ++i) {
    var toolbar = getToolbarAt(i);
    if (isCustomizableToolbar(toolbar)) {
      var toolbarMode = (localDefault && toolbar.hasAttribute("defaultmode")) ?
                        toolbar.getAttribute("defaultmode") :
                        aModeValue;
      setAttribute(toolbar, "mode", toolbarMode);
      gToolboxDocument.persist(toolbar.id, "mode");
    }
  }

  var iconSizeCheckbox = document.getElementById("smallicons");
  iconSizeCheckbox.disabled = aModeValue == "text";

  toolboxChanged();
}


function setAttribute(aElt, aAttr, aVal)
{
  if (aVal)
    aElt.setAttribute(aAttr, aVal);
  else
    aElt.removeAttribute(aAttr);
}

function isCustomizableToolbar(aElt)
{
  return aElt.localName == "toolbar" &&
         aElt.getAttribute("customizable") == "true";
}

function isSpecialItem(aElt)
{
  return aElt.localName == "toolbarseparator" ||
         aElt.localName == "toolbarspring" ||
         aElt.localName == "toolbarspacer";
}

function isToolbarItem(aElt)
{
  return aElt.localName == "toolbarbutton" ||
         aElt.localName == "toolbaritem" ||
         aElt.localName == "toolbarseparator" ||
         aElt.localName == "toolbarspring" ||
         aElt.localName == "toolbarspacer";
}

///////////////////////////////////////////////////////////////////////////
//// Drag and Drop observers

function onToolbarDragStart(aEvent)
{
  nsDragAndDrop.startDrag(aEvent, dragStartObserver);
}

function onToolbarDragOver(aEvent)
{
  nsDragAndDrop.dragOver(aEvent, toolbarDNDObserver);
}

function onToolbarDrop(aEvent)
{
  nsDragAndDrop.drop(aEvent, toolbarDNDObserver);
}

function onToolbarDragLeave(aEvent)
{
  if (gCurrentDragOverItem)
    setDragActive(gCurrentDragOverItem, false);
}

var dragStartObserver =
{
  onDragStart: function (aEvent, aXferData, aDragAction) {
    var documentId = gToolboxDocument.documentElement.id;
  
    var item = aEvent.target;
    while (item && item.localName != "toolbarpaletteitem")
      item = item.parentNode;

    item.setAttribute("dragactive", "true");
    
    aXferData.data = new TransferDataSet();
    var data = new TransferData();
    data.addDataForFlavour("text/toolbarwrapper-id/"+documentId, item.firstChild.id);
    aXferData.data.push(data);
    aDragAction.action = Components.interfaces.nsIDragService.DRAGDROP_ACTION_MOVE;
  }
}

var toolbarDNDObserver =
{
  onDragOver: function (aEvent, aFlavour, aDragSession)
  {
    var toolbar = aEvent.target;
    var dropTarget = aEvent.target;
    while (toolbar && toolbar.localName != "toolbar") {
      dropTarget = toolbar;
      toolbar = toolbar.parentNode;
    }

    var previousDragItem = gCurrentDragOverItem;

    // Make sure we are dragging over a customizable toolbar.
    if (!isCustomizableToolbar(toolbar)) {
      gCurrentDragOverItem = null;
      return;
    }

    if (dropTarget.localName == "toolbar") {
      gCurrentDragOverItem = dropTarget;
    } else {
      gCurrentDragOverItem = null;

      var direction = window.getComputedStyle(dropTarget.parentNode, null).direction;
      var dropTargetCenter = dropTarget.boxObject.x + (dropTarget.boxObject.width / 2);
      if (direction == "ltr")
        dragAfter = aEvent.clientX > dropTargetCenter;
      else
        dragAfter = aEvent.clientX < dropTargetCenter;
    
      if (dragAfter) {
        gCurrentDragOverItem = dropTarget.nextSibling;
        if (!gCurrentDragOverItem)
          gCurrentDragOverItem = toolbar;
      } else
        gCurrentDragOverItem = dropTarget;
    }    

    if (previousDragItem && gCurrentDragOverItem != previousDragItem) {
      setDragActive(previousDragItem, false);
    }

    setDragActive(gCurrentDragOverItem, true);

    aDragSession.canDrop = true;
  },
    
  onDrop: function (aEvent, aXferData, aDragSession)
  {
    if (!gCurrentDragOverItem)
      return;

    setDragActive(gCurrentDragOverItem, false);

    var draggedItemId = aXferData.data;
    if (gCurrentDragOverItem.id == draggedItemId)
      return;

    var toolbar = aEvent.target;
    while (toolbar.localName != "toolbar")
      toolbar = toolbar.parentNode;

    var draggedPaletteWrapper = document.getElementById("wrapper-"+draggedItemId);       
    if (!draggedPaletteWrapper) {
      // The wrapper has been dragged from the toolbar.
    
      // Get the wrapper from the toolbar document and make sure that
      // it isn't being dropped on itself.
      var wrapper = gToolboxDocument.getElementById("wrapper-"+draggedItemId);
      if (wrapper == gCurrentDragOverItem)
        return;

      // Don't allow static kids (e.g., the menubar) to move.
      if (wrapper.parentNode.firstPermanentChild && wrapper.parentNode.firstPermanentChild.id == wrapper.firstChild.id)
        return;
      if (wrapper.parentNode.lastPermanentChild && wrapper.parentNode.lastPermanentChild.id == wrapper.firstChild.id)
        return;

      // Remove the item from its place in the toolbar.
      wrapper.parentNode.removeChild(wrapper);

      // Determine which toolbar we are dropping on.
      var dropToolbar = null;
      if (gCurrentDragOverItem.localName == "toolbar")
        dropToolbar = gCurrentDragOverItem;
      else
        dropToolbar = gCurrentDragOverItem.parentNode;
  
      // Insert the item into the toolbar.
      if (gCurrentDragOverItem != dropToolbar)
        dropToolbar.insertBefore(wrapper, gCurrentDragOverItem);
      else
        dropToolbar.appendChild(wrapper);
    } else {
      // The item has been dragged from the palette

      // Create a new wrapper for the item. We don't know the id yet.
      var wrapper = createWrapper("", gToolboxDocument);
      var panel = document.getElementById("customizetabs").selectedPanel;

      if (draggedItemId != "separator" &&
          draggedItemId != "spring" &&
          draggedItemId != "spacer" &&
          gToolbox.palette.getElementsByAttribute("id", draggedItemId).length == 0)
      {
        if (panel.id == "mailboxestab") {
          var paletteBox = document.getElementById("mailbox-palette");
          var node = paletteBox.getElementsByAttribute("id", draggedItemId)[0];
          var foldertype = node.getAttribute("foldertype");
          var set = draggedItemId.split("/");
          var name = set[set.length-1];

          if (foldertype == "moveto-folder") {
            var newitem = createFolderMovetoButton(node.getAttribute("link"), name);
            gToolboxDocument.adoptNode(newitem);
            gToolbox.palette.appendChild(newitem);
          } else if (foldertype == "open-folder") {
            var newitem = createFolderOpenButton(node.getAttribute("link"), name);
            gToolboxDocument.adoptNode(newitem);
            gToolbox.palette.appendChild(newitem);
          }
        } else if (panel.id == "recipientstab") {
          var paletteBox = document.getElementById("recipient-palette");
          var node = paletteBox.getElementsByAttribute("id", draggedItemId)[0];
          var foldertype = node.getAttribute("foldertype");
          var set = draggedItemId.split("/");
          var name = node.getAttribute("label");
          var email = node.getAttribute("link");
          var buttontype = node.getAttribute("foldertype");

          var newitem = createRecipientButton(email, name, buttontype);
          gToolboxDocument.adoptNode(newitem);
          gToolbox.palette.appendChild(newitem);
        }
      }

      // Ask the toolbar to clone the item's template, place it inside the wrapper, and insert it in the toolbar.
      var newItem = toolbar.insertItem(draggedItemId, gCurrentDragOverItem == toolbar ? null : gCurrentDragOverItem, wrapper);

      // Prepare the item and wrapper to look good on the toolbar.
      cleanupItemForToolbar(newItem, wrapper);
      wrapper.id = "wrapper-"+newItem.id;
      wrapper.flex = newItem.flex;
    }

    gCurrentDragOverItem = null;

    toolboxChanged(true);
  },
    
  _flavourSet: null,

  getSupportedFlavours: function ()
  {
    if (!this._flavourSet) {
      this._flavourSet = new FlavourSet();
      var documentId = gToolboxDocument.documentElement.id;
      this._flavourSet.appendFlavour("text/toolbarwrapper-id/"+documentId);
    }
    return this._flavourSet;
  }
}

var paletteDNDObserver =
{
  onDragOver: function (aEvent, aFlavour, aDragSession)
  {
    aDragSession.canDrop = true;
  },

  onDrop: function(aEvent, aXferData, aDragSession)
  {
    var itemId = aXferData.data;
    
    var wrapper = gToolboxDocument.getElementById("wrapper-"+itemId);
    if (wrapper) {
      // Don't allow static kids (e.g., the menubar) to move.
      if (wrapper.parentNode.firstPermanentChild && wrapper.parentNode.firstPermanentChild.id == wrapper.firstChild.id)
        return;
      if (wrapper.parentNode.lastPermanentChild && wrapper.parentNode.lastPermanentChild.id == wrapper.firstChild.id)
        return;

      var wrapperType = wrapper.getAttribute("type");
      if (wrapperType != "separator" &&
          wrapperType != "spacer" &&
          wrapperType != "spring") {
        //appendPaletteItem(document.importNode(wrapper.firstChild, true), aEvent.currentTarget);
        gToolbox.palette.appendChild(wrapper.firstChild);
      }

      // The item was dragged out of the toolbar.
      wrapper.parentNode.removeChild(wrapper);

      toolboxChanged(true);
    }
  },

  _flavourSet: null,

  getSupportedFlavours: function ()
  {
    if (!this._flavourSet) {
      this._flavourSet = new FlavourSet();
      var documentId = gToolboxDocument.documentElement.id;
      this._flavourSet.appendFlavour("text/toolbarwrapper-id/"+documentId);
    }
    return this._flavourSet;
  }
}

function createMenubarTree(node, curparent)
{
  var datasources = node.getAttribute("datasources");
  if (datasources)
  {
    // Not possible to resolve menu populated via templates and RDF
    return;
  }
  var label = node.getAttribute("label");
  var newparent;
  if (label) {
    var treeitem = document.createElement("treeitem");
    var treerow = document.createElement("treerow");
    var treecell = document.createElement("treecell");
    var treechildren = document.createElement("treechildren");
    treecell.setAttribute("label", label);
    treeitem.setAttribute("container", "true");
    treerow.appendChild(treecell);
    treeitem.appendChild(treerow);
    treeitem.appendChild(treechildren);
    curparent.appendChild(treeitem);
    newparent = treechildren;
  } else {
    newparent = curparent;
  }
  var hidden;
  var toolbaritem;
  var child = node.childNodes;
  for (var j=0; j < child.length; j++) {
    if (child[j].nodeType == Node.TEXT_NODE)
      continue;
    hidden = child[j].getAttribute("hidden");
    if (hidden == "true")
      continue;
    toolbaritem = child[j].getAttribute("toolbaritem");
    if (toolbaritem == "not-supported")
      continue;

    if (toolbaritem) {
      label = child[j].getAttribute("label");
      // strip off trailing ellipsis
      while (label.length && (label[label.length - 1] == '…' || label[label.length - 1] == '.'))
        label = label.slice(0, -1);
      var treeitem = document.createElement("treeitem");
      var treerow = document.createElement("treerow");
      var treecell = document.createElement("treecell");
      treecell.setAttribute("label", label);
      treecell.setAttribute("toolbaritem", toolbaritem);
      treerow.appendChild(treecell);
      treeitem.appendChild(treerow);
      newparent.appendChild(treeitem);
    } else if (child[j].hasChildNodes()) {
      createMenubarTree(child[j], newparent);
    }
  }
  if (newparent != curparent && newparent.childElementCount == 0)
    curparent.removeChild(newparent.parentNode);
}

function onMenuItemSelect()
{
  var elem = gMenuItemTree.view.getItemAtIndex(gMenuItemTree.currentIndex);
  var cell = elem.firstChild.firstChild;
  var label = cell.getAttribute("label");
  var toolbaritem = cell.getAttribute("toolbaritem");
  var descId = "menuItemDescription";

  if (toolbaritem)
    buildPalette('general-palette', toolbaritem, descId);
  else
    notSupported('general-palette', descId);
}

function onMboxSelect()
{
  var elem = gMboxTree.view.getItemAtIndex(gMboxTree.currentIndex);
  var cell = elem.firstChild.firstChild;
  var label = cell.getAttribute("label");
  var allowFileOnServer = cell.getAttribute("allowFileOnServer");
  var allowFile = cell.getAttribute("allowFile");
  buildMailboxPalette(elem.id, label, (allowFileOnServer == "false" || allowFile=="false"));
}
 
function onRecipientSelect()
{
  var index = GetSelectedCardIndex();
  if (index >= 0) {
    var card = gAbView.getCardFromRow(index);
    var email = card.primaryEmail;
    var name = card.displayName;
    if (!name)
      name = email;
    buildRecipientsPalette(email, name);
  }
}

function notSupported(mypalette, descId)
{
  // Use a non-breaking space for the "empty" description so that the items
  // above the description don't resize when switching between an item with
  // a description and one without.  For some reason a regular space doesn't
  // work (maybe it gets stripped out?).
  var linetext = document.createTextNode("\xA0");
  var desc = document.getElementById(descId);
  while (desc.firstChild)
    desc.removeChild(desc.firstChild);
  desc.appendChild(linetext);

  var paletteBox = document.getElementById(mypalette);
  while (paletteBox.lastChild)
    paletteBox.removeChild(paletteBox.lastChild);
}

function setDescText(descId, text)
{
  var linetext = document.createTextNode(text);
  var desc = document.getElementById(descId);
  while (desc.firstChild)
    desc.removeChild(desc.firstChild);
  desc.appendChild(linetext);
}

function goUpdateCommand(unused)
{
  dump("called goUpdateCommand: "+unused+"\n");
}

function ClearCardViewPane()
{
  // Stub to prevent unresolved function
}

